/*         ______   ___    ___ 
 *        /\  _  \ /\_ \  /\_ \ 
 *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___ 
 *         \ \  __ \ \ \ \  \ \ \   /'__`\ /'_ `\/\`'__\/ __`\
 *          \ \ \/\ \ \_\ \_ \_\ \_/\  __//\ \L\ \ \ \//\ \L\ \
 *           \ \_\ \_\/\____\/\____\ \____\ \____ \ \_\\ \____/
 *            \/_/\/_/\/____/\/____/\/____/\/___L\ \/_/ \/___/
 *                                           /\____/
 *                                           \_/__/
 *
 *      File I/O.
 *
 *      By Shawn Hargreaves.
 *
 *      _pack_fdopen() and related modifications by Annie Testes.
 *
 *      Evert Glebbeek added the support for relative filenames:
 *      make_absolute_filename(), make_relative_filename() and
 *      is_relative_filename().
 *
 *      Peter Wang added support for packfile vtables.
 *
 *      See readme.txt for copyright information.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>
#include <allegro5/allegro.h>

#include "t3f/debug.h"

#ifndef ALLEGRO_MPW
   #include <sys/stat.h>
#endif

#ifdef ALLEGRO_UNIX
   #include <pwd.h>                 /* for tilde expansion */
#endif

#include "file.h"
#include "lzss.h"

#ifndef O_BINARY
   #define O_BINARY  0
#endif

/* permissions to use when opening files */
#ifndef ALLEGRO_MPW

/* some OSes have no concept of "group" and "other" */
#ifndef S_IRGRP
   #define S_IRGRP   0
   #define S_IWGRP   0
#endif
#ifndef S_IROTH
   #define S_IROTH   0
   #define S_IWOTH   0
#endif

#define OPEN_PERMS   (S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH)

#endif /* !ALLEGRO_MPW */



static char the_password[256] = "";

int _packfile_filesize = 0;
int _packfile_datasize = 0;

int _packfile_type = 0;

static PACKFILE_VTABLE normal_vtable;

#define FA_DAT_FLAGS  (FA_RDONLY | FA_ARCH)


typedef struct RESOURCE_PATH
{
   int priority;
   char path[1024];
   struct RESOURCE_PATH *next;
} RESOURCE_PATH;


/***************************************************
 ******************** Packfiles ********************
 ***************************************************/


/* packfile_password:
 *  Sets the password to be used by all future read/write operations.
 *  This only affects "normal" PACKFILEs, i.e. ones not using user-supplied
 *  packfile vtables.
 */
void packfile_password(const char *password)
{
   int i = 0;
   int c;

   if (password) {
      while ((c = password[i]) != 0) {
	 the_password[i++] = c;
	 if (i >= (int)sizeof(the_password)-1)
	    break;
      }
   }

   the_password[i] = 0;
}



/* encrypt_id:
 *  Helper for encrypting magic numbers, using the current password.
 */
static long encrypt_id(long x, int new_format)
{
   long mask = 0;
   int i, pos;

   if (the_password[0]) {
      for (i=0; the_password[i]; i++)
	 mask ^= ((long)the_password[i] << ((i&3) * 8));

      for (i=0, pos=0; i<4; i++) {
	 mask ^= (long)the_password[pos++] << (24-i*8);
	 if (!the_password[pos])
	    pos = 0;
      }

      if (new_format)
	 mask ^= 42;
   }

   return x ^ mask;
}



/* clone_password:
 *  Sets up a local password string for use by this packfile.
 */
static int clone_password(PACKFILE *f)
{
   if (the_password[0]) {
      if ((f->normal.passdata = malloc(strlen(the_password)+1)) == NULL) {
	 return 0;
      }
      strncpy(f->normal.passdata, the_password, strlen(the_password)+1);
      f->normal.passpos = f->normal.passdata;
   }
   else {
      f->normal.passpos = NULL;
      f->normal.passdata = NULL;
   }

   return 1;
}



/* create_packfile:
 *  Helper function for creating a PACKFILE structure.
 */
static PACKFILE *create_packfile(int is_normal_packfile)
{
   PACKFILE *f;

   if (is_normal_packfile)
      f = malloc(sizeof(PACKFILE));
   else
      f = malloc(sizeof(PACKFILE) - sizeof(struct _al_normal_packfile_details));

   if (f == NULL) {
      return NULL;
   }

   if (!is_normal_packfile) {
      f->vtable = NULL;
      f->userdata = NULL;
      f->is_normal_packfile = 0;
   }
   else {
      f->vtable = &normal_vtable;
      f->userdata = f;
      f->is_normal_packfile = 1;

      f->normal.buf_pos = f->normal.buf;
      f->normal.flags = 0;
      f->normal.buf_size = 0;
      f->normal.filename = NULL;
      f->normal.passdata = NULL;
      f->normal.passpos = NULL;
      f->normal.parent = NULL;
      f->normal.pack_data = NULL;
      f->normal.unpack_data = NULL;
      f->normal.todo = 0;
      f->normal.pos = 0;
   }

   return f;
}



/* free_packfile:
 *  Helper function for freeing the PACKFILE struct.
 */
static void free_packfile(PACKFILE *f)
{
   if (f) {
      /* These are no longer the responsibility of this function, but
       * these assertions help catch instances of old code which still
       * rely on the old behaviour.
       */
      if (f->is_normal_packfile) {
      }

      free(f);
   }
}



/* _pack_fdopen:
 *  Converts the given file descriptor into a PACKFILE. The mode can have
 *  the same values as for pack_fopen() and must be compatible with the
 *  mode of the file descriptor. Unlike the libc fdopen(), pack_fdopen()
 *  is unable to convert an already partially read or written file (i.e.
 *  the file offset must be 0).
 *  On success, it returns a pointer to a file structure, and on error it
 *  returns NULL and stores an error code in errno. An attempt to read
 *  a normal file in packed mode will cause errno to be set to EDOM.
 */
PACKFILE *_pack_fpopen(const char * fn, ALLEGRO_FILE * fp, const char *mode)
{
	PACKFILE *f, *f2;
	long header = 0;
	int c;

	if((f = create_packfile(1)) == NULL)
	{
		return NULL;
	}

	while ((c = *(mode++)) != 0)
	{
		switch (c)
		{
			case 'r': case 'R': f->normal.flags &= ~PACKFILE_FLAG_WRITE; break;
			case 'w': case 'W': f->normal.flags |= PACKFILE_FLAG_WRITE; break;
			case 'p': case 'P': f->normal.flags |= PACKFILE_FLAG_PACK; break;
			case '!': f->normal.flags &= ~PACKFILE_FLAG_PACK; header = 1; break;
		}
	}

	if (f->normal.flags & PACKFILE_FLAG_WRITE)
	{
		if (f->normal.flags & PACKFILE_FLAG_PACK)
		{
			/* write a packed file */
			f->normal.pack_data = create_lzss_pack_data();

			if (!f->normal.pack_data)
			{
				free_packfile(f);
				return NULL;
			}

			if ((f->normal.parent = _pack_fpopen(fn, fp, "wb")) == NULL)
			{
				free_lzss_pack_data(f->normal.pack_data);
				f->normal.pack_data = NULL;
				free_packfile(f);
				return NULL;
			}

			pack_mputl(encrypt_id(F_PACK_MAGIC, 1), f->normal.parent);

			f->normal.todo = 4;
		}
		else
		{
			/* write a 'real' file */
			if (!clone_password(f))
			{
				free_packfile(f);
				return NULL;
			}

			f->normal.fp = fp;
			f->normal.todo = 0;

			if (header)
			{
				pack_mputl(encrypt_id(F_NOPACK_MAGIC, 1), f);
			}
		}
	}
	else
	{ 
		if (f->normal.flags & PACKFILE_FLAG_PACK)
		{
			/* read a packed file */
			f->normal.unpack_data = create_lzss_unpack_data();

			if (!f->normal.unpack_data)
			{
				free_packfile(f);
				return NULL;
			}

			if ((f->normal.parent = _pack_fpopen(fn, fp, "rb")) == NULL)
			{
				free_lzss_unpack_data(f->normal.unpack_data);
				f->normal.unpack_data = NULL;
				free_packfile(f);
				return NULL;
			}

			header = pack_mgetl(f->normal.parent);

			if ((f->normal.parent->normal.passpos) &&
			   ((header == encrypt_id(F_PACK_MAGIC, 0)) ||
			   (header == encrypt_id(F_NOPACK_MAGIC, 0))))
			{
				/* duplicate the file descriptor */
				ALLEGRO_FILE * fp2 = al_fopen(fn, mode);
  
				if (!fp2)
				{
					pack_fclose(f->normal.parent);
					free_packfile(f);
					return NULL;
				}
  
				/* close the parent file (logically, not physically) */
				pack_fclose(f->normal.parent);
  
				/* backward compatibility mode */
				if (!clone_password(f))
				{
					free_packfile(f);
					return NULL;
				}
  
				f->normal.flags |= PACKFILE_FLAG_OLD_CRYPT;
  
				/* re-open the parent file */
				al_fseek(fp2, 0, SEEK_SET);
  
				if ((f->normal.parent = _pack_fpopen(fn, fp2, F_READ)) == NULL)
				{
					free_packfile(f);
					return NULL;
				}
  
				f->normal.parent->normal.flags |= PACKFILE_FLAG_OLD_CRYPT;
  
				pack_mgetl(f->normal.parent);
  
				if (header == encrypt_id(F_PACK_MAGIC, 0))
				{
					header = encrypt_id(F_PACK_MAGIC, 1);
				}
				else
				{
					header = encrypt_id(F_NOPACK_MAGIC, 1);
				}
			}

			if (header == encrypt_id(F_PACK_MAGIC, 1))
			{
				f->normal.todo = LONG_MAX;
			}
			else if (header == encrypt_id(F_NOPACK_MAGIC, 1))
			{
				f2 = f->normal.parent;
				free_lzss_unpack_data(f->normal.unpack_data);
				f->normal.unpack_data = NULL;
				free_packfile(f);
				return f2;
			}
			else
			{
				pack_fclose(f->normal.parent);
				free_lzss_unpack_data(f->normal.unpack_data);
				f->normal.unpack_data = NULL;
				free_packfile(f);
				return NULL;
			}
		}
		else
		{
			/* read a 'real' file */
			f->normal.todo = al_fsize(fp);  /* size of the file */
			if (f->normal.todo < 0)
			{
				free_packfile(f);
				return NULL;
			}

			al_fseek(fp, 0, SEEK_SET);

			if (!clone_password(f))
			{
				free_packfile(f);
				return NULL;
			}

			f->normal.fp = fp;
		}
	}

	return f;
}



/* pack_fopen:
 *  Opens a file according to mode, which may contain any of the flags:
 *  'r': open file for reading.
 *  'w': open file for writing, overwriting any existing data.
 *  'p': open file in 'packed' mode. Data will be compressed as it is
 *       written to the file, and automatically uncompressed during read
 *       operations. Files created in this mode will produce garbage if
 *       they are read without this flag being set.
 *  '!': open file for writing in normal, unpacked mode, but add the value
 *       F_NOPACK_MAGIC to the start of the file, so that it can be opened
 *       in packed mode and Allegro will automatically detect that the
 *       data does not need to be decompressed.
 *
 *  Instead of these flags, one of the constants F_READ, F_WRITE,
 *  F_READ_PACKED, F_WRITE_PACKED or F_WRITE_NOPACK may be used as the second 
 *  argument to fopen().
 *
 *  On success, fopen() returns a pointer to a file structure, and on error
 *  it returns NULL and stores an error code in errno. An attempt to read a 
 *  normal file in packed mode will cause errno to be set to EDOM.
 */
PACKFILE *pack_fopen(const char *filename, const char *mode)
{
   ALLEGRO_FILE * fp;

   _packfile_type = 0;

#ifndef ALLEGRO_MPW
   if (strpbrk(mode, "wW"))  /* write mode? */
      fp = al_fopen(filename, "wb");
   else
      fp = al_fopen(filename, "rb");
#else
   if (strpbrk(mode, "wW"))  /* write mode? */
      fp = al_fopen(filename, "wb");
   else
      fp = al_fopen(filename, "rb");
#endif

   if (!fp) {
      return NULL;
   }

   return _pack_fpopen(filename, fp, mode);
}



/* pack_fopen_vtable:
 *  Creates a new packfile structure that uses the functions specified in
 *  the vtable instead of the standard functions.  On success, it returns a
 *  pointer to a file structure, and on error it returns NULL and
 *  stores an error code in errno.
 *
 *  The vtable and userdata must remain available for the lifetime of the
 *  created packfile.
 *
 *  Opening chunks using pack_fopen_chunk() on top of the returned packfile
 *  is not possible at this time.
 *
 *  packfile_password() does not have any effect on packfiles opened
 *  with pack_fopen_vtable().
 */
PACKFILE *pack_fopen_vtable(const PACKFILE_VTABLE *vtable, void *userdata)
{
   PACKFILE *f;

   if ((f = create_packfile(0)) == NULL)
      return NULL;

   f->vtable = vtable;
   f->userdata = userdata;

   return f;
}



/* pack_fclose:
 *  Closes a file after it has been read or written.
 *  Returns zero on success. On error it returns an error code which is
 *  also stored in errno. This function can fail only when writing to
 *  files: if the file was opened in read mode it will always succeed.
 */
int pack_fclose(PACKFILE *f)
{
   int ret;

   if (!f)
      return 0;

   ret = f->vtable->pf_fclose(f->userdata);

   free_packfile(f);

   return ret;
}



/* pack_fopen_chunk: 
 *  Opens a sub-chunk of the specified file, for reading or writing depending
 *  on the type of the file. The returned file pointer describes the sub
 *  chunk, and replaces the original file, which will no longer be valid.
 *  When writing to a chunk file, data is sent to the original file, but
 *  is prefixed with two length counts (32 bit, big-endian). For uncompressed 
 *  chunks these will both be set to the length of the data in the chunk.
 *  For compressed chunks, created by setting the pack flag, the first will
 *  contain the raw size of the chunk, and the second will be the negative
 *  size of the uncompressed data. When reading chunks, the pack flag is
 *  ignored, and the compression type is detected from the sign of the
 *  second size value. The file structure used to read chunks checks the
 *  chunk size, and will return EOF if you try to read past the end of
 *  the chunk. If you don't read all of the chunk data, when you call
 *  pack_fclose_chunk(), the parent file will advance past the unused data.
 *  When you have finished reading or writing a chunk, you should call
 *  pack_fclose_chunk() to return to your original file.
 */
PACKFILE *pack_fopen_chunk(PACKFILE *f, int pack)
{
   PACKFILE *chunk;
   char *name = "abc";

   /* unsupported */
   if (!f->is_normal_packfile) {
      return NULL;
   }

   if (f->normal.flags & PACKFILE_FLAG_WRITE) {

      /* write a sub-chunk */ 
      ALLEGRO_FILE * tmp_fp = NULL;
      tmp_fp = al_make_temp_file("pfXXXX", NULL);
      if (!tmp_fp) {
         return NULL;
      }

      chunk = _pack_fpopen(f->normal.filename, tmp_fp, (pack ? F_WRITE_PACKED : F_WRITE_NOPACK));

      if (chunk) {
         chunk->normal.filename = strdup(name);

	 if (pack)
	    chunk->normal.parent->normal.parent = f;
	 else
	    chunk->normal.parent = f;

	 chunk->normal.flags |= PACKFILE_FLAG_CHUNK;
      }
      
   }
   else {
      /* read a sub-chunk */
      _packfile_filesize = pack_mgetl(f);
      _packfile_datasize = pack_mgetl(f);

      if ((chunk = create_packfile(1)) == NULL)
         return NULL;

      chunk->normal.flags = PACKFILE_FLAG_CHUNK;
      chunk->normal.parent = f;

      if (f->normal.flags & PACKFILE_FLAG_OLD_CRYPT) {
	 /* backward compatibility mode */
	 if (f->normal.passdata) {
	    if ((chunk->normal.passdata = malloc(strlen(f->normal.passdata)+1)) == NULL) {
	       free(chunk);
	       return NULL;
	    }
	    strncpy(chunk->normal.passdata, f->normal.passdata, strlen(f->normal.passdata)+1);
	    chunk->normal.passpos = chunk->normal.passdata + (long)f->normal.passpos - (long)f->normal.passdata;
	    f->normal.passpos = f->normal.passdata;
	 }
	 chunk->normal.flags |= PACKFILE_FLAG_OLD_CRYPT;
      }

      if (_packfile_datasize < 0) {
	 /* read a packed chunk */
         chunk->normal.unpack_data = create_lzss_unpack_data();

	 if (!chunk->normal.unpack_data) {
            free_packfile(chunk);
	    return NULL;
	 }

	 _packfile_datasize = -_packfile_datasize;
	 chunk->normal.todo = _packfile_datasize;
	 chunk->normal.flags |= PACKFILE_FLAG_PACK;
      }
      else {
	 /* read an uncompressed chunk */
	 chunk->normal.todo = _packfile_datasize;
      }
   }

   return chunk;
}



/* pack_fclose_chunk:
 *  Call after reading or writing a sub-chunk. This closes the chunk file,
 *  and returns a pointer to the original file structure (the one you
 *  passed to pack_fopen_chunk()), to allow you to read or write data 
 *  after the chunk. If an error occurs, returns NULL and sets errno.
 */
PACKFILE *pack_fclose_chunk(PACKFILE *f)
{
   PACKFILE *parent;
   PACKFILE *tmp;
   char *name;
   int header, c;

   /* unsupported */
   if (!f->is_normal_packfile) {
      return NULL;
   }

   parent = f->normal.parent;
   name = f->normal.filename;

   if (f->normal.flags & PACKFILE_FLAG_WRITE) {
      /* finish writing a chunk */
      ALLEGRO_FILE * fp;

      /* duplicate the file descriptor to create a readable pack file,
       * the file descriptor must have been opened in read/write mode
       */
      if (f->normal.flags & PACKFILE_FLAG_PACK)
         fp = al_fopen(f->normal.parent->normal.filename, f->normal.parent->normal.mode);
      else
         fp = al_fopen(f->normal.filename, f->normal.mode);

      if (fp) {
         return NULL;
      }

      _packfile_datasize = f->normal.todo + f->normal.buf_size - 4;

      if (f->normal.flags & PACKFILE_FLAG_PACK) {
	 parent = parent->normal.parent;
	 f->normal.parent->normal.parent = NULL;
      }
      else
	 f->normal.parent = NULL;

      /* close the writeable temp file, it isn't physically closed
       * because the descriptor has been duplicated
       */
      f->normal.flags &= ~PACKFILE_FLAG_CHUNK;
      pack_fclose(f);

      al_fseek(fp, 0, SEEK_SET);

      /* create a readable pack file */
      tmp = _pack_fpopen(f->normal.filename, fp, "rb");
      if (!tmp)
         return NULL;

      _packfile_filesize = tmp->normal.todo - 4;

      header = pack_mgetl(tmp);

      pack_mputl(_packfile_filesize, parent);

      if (header == encrypt_id(F_PACK_MAGIC, 1))
	 pack_mputl(-_packfile_datasize, parent);
      else
	 pack_mputl(_packfile_datasize, parent);

      while ((c = pack_getc(tmp)) != EOF)
	 pack_putc(c, parent);

      pack_fclose(tmp);

      al_remove_filename(name);
      free(name);
   }
   else {
      /* finish reading a chunk */
      while (f->normal.todo > 0)
	 pack_getc(f);

      if (f->normal.unpack_data) {
	 free_lzss_unpack_data(f->normal.unpack_data);
	 f->normal.unpack_data = NULL;
      }

      if ((f->normal.passpos) && (f->normal.flags & PACKFILE_FLAG_OLD_CRYPT))
	 parent->normal.passpos = parent->normal.passdata + (long)f->normal.passpos - (long)f->normal.passdata;

      free_packfile(f);
   }

   return parent;
}



/* pack_fseek:
 *  Like the stdio fseek() function, but only supports forward seeks 
 *  relative to the current file position.
 */
int pack_fseek(PACKFILE *f, int offset)
{
   return f->vtable->pf_fseek(f->userdata, offset);
}



/* pack_getc:
 *  Returns the next character from the stream f, or EOF if the end of the
 *  file has been reached.
 */
int pack_getc(PACKFILE *f)
{
	int c = f->vtable->pf_getc(f->userdata);
	f->normal.pos++;
	return c;
}



/* pack_putc:
 *  Puts a character in the stream f.
 */
int pack_putc(int c, PACKFILE *f)
{
	int ret = f->vtable->pf_putc(c, f->userdata);
	f->normal.pos++;
	return ret;
}



/* pack_feof:
 *  pack_feof() returns nonzero as soon as you reach the end of the file. It 
 *  does not wait for you to attempt to read beyond the end of the file,
 *  contrary to the ISO C feof() function.
 */
int pack_feof(PACKFILE *f)
{
   return f->vtable->pf_feof(f->userdata);
}



/* pack_ferror:
 *  Returns nonzero if the error indicator for the stream is set, indicating
 *  that an error has occurred during a previous operation on the stream.
 */
int pack_ferror(PACKFILE *f)
{
   return f->vtable->pf_ferror(f->userdata);
}



/* pack_igetw:
 *  Reads a 16 bit word from a file, using intel byte ordering.
 */
int pack_igetw(PACKFILE *f)
{
   int b1, b2;

   if ((b1 = pack_getc(f)) != EOF)
      if ((b2 = pack_getc(f)) != EOF)
	 return ((b2 << 8) | b1);

   return EOF;
}



/* pack_igetl:
 *  Reads a 32 bit long from a file, using intel byte ordering.
 */
long pack_igetl(PACKFILE *f)
{
   int b1, b2, b3, b4;

   if ((b1 = pack_getc(f)) != EOF)
      if ((b2 = pack_getc(f)) != EOF)
	 if ((b3 = pack_getc(f)) != EOF)
	    if ((b4 = pack_getc(f)) != EOF)
	       return (((long)b4 << 24) | ((long)b3 << 16) |
		       ((long)b2 << 8) | (long)b1);

   return EOF;
}



/* pack_iputw:
 *  Writes a 16 bit int to a file, using intel byte ordering.
 */
int pack_iputw(int w, PACKFILE *f)
{
   int b1, b2;

   b1 = (w & 0xFF00) >> 8;
   b2 = w & 0x00FF;

   if (pack_putc(b2,f)==b2)
      if (pack_putc(b1,f)==b1)
	 return w;

   return EOF;
}



/* pack_iputl:
 *  Writes a 32 bit long to a file, using intel byte ordering.
 */
long pack_iputl(long l, PACKFILE *f)
{
   int b1, b2, b3, b4;

   b1 = (int)((l & 0xFF000000L) >> 24);
   b2 = (int)((l & 0x00FF0000L) >> 16);
   b3 = (int)((l & 0x0000FF00L) >> 8);
   b4 = (int)l & 0x00FF;

   if (pack_putc(b4,f)==b4)
      if (pack_putc(b3,f)==b3)
	 if (pack_putc(b2,f)==b2)
	    if (pack_putc(b1,f)==b1)
	       return l;

   return EOF;
}



/* pack_mgetw:
 *  Reads a 16 bit int from a file, using motorola byte-ordering.
 */
int pack_mgetw(PACKFILE *f)
{
   int b1, b2;

   if ((b1 = pack_getc(f)) != EOF)
      if ((b2 = pack_getc(f)) != EOF)
	 return ((b1 << 8) | b2);

   return EOF;
}



/* pack_mgetl:
 *  Reads a 32 bit long from a file, using motorola byte-ordering.
 */
long pack_mgetl(PACKFILE *f)
{
   int b1, b2, b3, b4;

   if ((b1 = pack_getc(f)) != EOF)
      if ((b2 = pack_getc(f)) != EOF)
	 if ((b3 = pack_getc(f)) != EOF)
	    if ((b4 = pack_getc(f)) != EOF)
	       return (((long)b1 << 24) | ((long)b2 << 16) |
		       ((long)b3 << 8) | (long)b4);

   return EOF;
}



/* pack_mputw:
 *  Writes a 16 bit int to a file, using motorola byte-ordering.
 */
int pack_mputw(int w, PACKFILE *f)
{
   int b1, b2;

   b1 = (w & 0xFF00) >> 8;
   b2 = w & 0x00FF;

   if (pack_putc(b1,f)==b1)
      if (pack_putc(b2,f)==b2)
	 return w;

   return EOF;
}



/* pack_mputl:
 *  Writes a 32 bit long to a file, using motorola byte-ordering.
 */
long pack_mputl(long l, PACKFILE *f)
{
   int b1, b2, b3, b4;

   b1 = (int)((l & 0xFF000000L) >> 24);
   b2 = (int)((l & 0x00FF0000L) >> 16);
   b3 = (int)((l & 0x0000FF00L) >> 8);
   b4 = (int)l & 0x00FF;

   if (pack_putc(b1,f)==b1)
      if (pack_putc(b2,f)==b2)
	 if (pack_putc(b3,f)==b3)
	    if (pack_putc(b4,f)==b4)
	       return l;

   return EOF;
}



/* pack_fread:
 *  Reads n bytes from f and stores them at memory location p. Returns the 
 *  number of items read, which will be less than n if EOF is reached or an 
 *  error occurs. Error codes are stored in errno.
 */
long pack_fread(void *p, long n, PACKFILE *f)
{
	int ret = f->vtable->pf_fread(p, n, f->userdata);
	f->normal.pos += n;
	return ret;
}



/* pack_fwrite:
 *  Writes n bytes to the file f from memory location p. Returns the number 
 *  of items written, which will be less than n if an error occurs. Error 
 *  codes are stored in errno.
 */
long pack_fwrite(const void *p, long n, PACKFILE *f)
{
	int ret = f->vtable->pf_fwrite(p, n, f->userdata);
	f->normal.pos += n;
	return ret;
}



/* pack_ungetc:
 *  Puts a character back in the file's input buffer. It only works
 *  for characters just fetched by pack_getc and, like ungetc, only a
 *  single push back is guaranteed.
 */
int pack_ungetc(int c, PACKFILE *f)
{
	int ret = f->vtable->pf_ungetc(c, f->userdata);
	f->normal.pos--;
	return ret;
}



/* pack_fgets:
 *  Reads a line from a text file, storing it at location p. Stops when a
 *  linefeed is encountered, or max bytes have been read. Returns a pointer
 *  to where it stored the text, or NULL on error. The end of line is
 *  handled by detecting optional '\r' characters optionally followed 
 *  by '\n' characters. This supports CR-LF (DOS/Windows), LF (Unix), and
 *  CR (Mac) formats.
 */
char *pack_fgets(char *p, int max, PACKFILE *f)
{
   char *pmax, *orig_p = p;
   int c;
   pmax = p+max - 1;

   if ((c = pack_getc(f)) == EOF) {
      if (1 <= max)
         *p = 0;
      return NULL;
   }

   do {

      if (c == '\r' || c == '\n') {
	 /* Technically we should check there's space in the buffer, and if so,
	  * add a \n.  But pack_fgets has never done this. */
	 if (c == '\r') {
	    /* eat the following \n, if any */
	    c = pack_getc(f);
	    if ((c != '\n') && (c != EOF))
	       pack_ungetc(c, f);
	 }
	 break;
      }

      /* is there room in the buffer? */
      if (1 > pmax - p) {
	 pack_ungetc(c, f);
	 c = '\0';
	 break;
      }

      /* write the character */
      *p = c;
   }
   while ((c = pack_getc(f)) != EOF);

   /* terminate the string */
   *p = 0;

   if (c == '\0')
      return NULL;

   return orig_p; /* p has changed */
}



/* pack_fputs:
 *  Writes a string to a text file, returning zero on success, -1 on error.
 *  The input string is converted from the current text encoding format
 *  to UTF-8 before writing. Newline characters (\n) are written as \r\n
 *  on DOS and Windows platforms.
 */
int pack_fputs(const char *p, PACKFILE *f)
{
   char *buf, *s;
   int bufsize;

   bufsize = strlen(p) + 1;
   buf = malloc(bufsize);
   if (!buf)
      return -1;

   s = (char *)p;

   while (*s) {
      #if (defined ALLEGRO_DOS) || (defined ALLEGRO_WINDOWS)
	 if (*s == '\n')
	    pack_putc('\r', f);
      #endif

      pack_putc(*s, f);
      s++;
   }

   free(buf);

   return 0;
}



/* pack_get_userdata:
 *  Returns the userdata field of packfiles using user-defined vtables.
 */
void *pack_get_userdata(PACKFILE *f)
{
   return f->userdata;
}



/***************************************************
 ************ "Normal" packfile vtable *************
 ***************************************************

   Ideally this would be the only section which knows about the details
   of "normal" packfiles. However, this ideal is still being violated in
   many places (partly due to the API, and partly because it would be
   quite a lot of work to move the _al_normal_packfile_details field
   into the userdata field of the PACKFILE structure.
*/


static int normal_fclose(void *_f);
static int normal_getc(void *_f);
static int normal_ungetc(int ch, void *_f);
static int normal_putc(int c, void *_f);
static long normal_fread(void *p, long n, void *_f);
static long normal_fwrite(const void *p, long n, void *_f);
static int normal_fseek(void *_f, int offset);
static int normal_feof(void *_f);
static int normal_ferror(void *_f);

static int normal_refill_buffer(PACKFILE *f);
static int normal_flush_buffer(PACKFILE *f, int last);



static PACKFILE_VTABLE normal_vtable =
{
   normal_fclose,
   normal_getc,
   normal_ungetc,
   normal_fread,
   normal_putc,
   normal_fwrite,
   normal_fseek,
   normal_feof,
   normal_ferror
};



static int normal_fclose(void *_f)
{
   PACKFILE *f = _f;
   int ret;

   if (f->normal.flags & PACKFILE_FLAG_WRITE) {
      if (f->normal.flags & PACKFILE_FLAG_CHUNK) {
	 f = pack_fclose_chunk(_f);
	 if (!f)
	    return -1;

	 return pack_fclose(f);
      }

      normal_flush_buffer(f, 1);
   }

   if (f->normal.parent) {
      ret = pack_fclose(f->normal.parent);
   }
   else {
      ret = al_fclose(f->normal.fp);
   }

   if (f->normal.pack_data) {
      free_lzss_pack_data(f->normal.pack_data);
      f->normal.pack_data = NULL;
   }

   if (f->normal.unpack_data) {
      free_lzss_unpack_data(f->normal.unpack_data);
      f->normal.unpack_data = NULL;
   }

   if (f->normal.passdata) {
      free(f->normal.passdata);
      f->normal.passdata = NULL;
      f->normal.passpos = NULL;
   }

   return ret;
}



/* normal_no_more_input:
 *  Return non-zero if the number of bytes remaining in the file (todo) is
 *  less than or equal to zero.
 *
 *  However, there is a special case.  If we are reading from a LZSS
 *  compressed file, the latest call to lzss_read() may have suspended while
 *  writing out a sequence of bytes due to the output buffer being too small.
 *  In that case the `todo' count would be decremented (possibly to zero),
 *  but the output isn't yet completely written out.
 */
static INLINE int normal_no_more_input(PACKFILE *f)
{
   /* see normal_refill_buffer() to see when lzss_read() is called */
   if (f->normal.parent && (f->normal.flags & PACKFILE_FLAG_PACK) &&
       _al_lzss_incomplete_state(f->normal.unpack_data))
      return 0;

   return (f->normal.todo <= 0);
}



static int normal_getc(void *_f)
{
   PACKFILE *f = _f;

   f->normal.buf_size--;
   if (f->normal.buf_size > 0)
      return *(f->normal.buf_pos++);

   if (f->normal.buf_size == 0) {
      if (normal_no_more_input(f))
	 f->normal.flags |= PACKFILE_FLAG_EOF;
      return *(f->normal.buf_pos++);
   }

   return normal_refill_buffer(f);
}



static int normal_ungetc(int c, void *_f)
{
   PACKFILE *f = _f;

   if (f->normal.buf_pos == f->normal.buf) {
      return EOF;
   }
   else {
      *(--f->normal.buf_pos) = (unsigned char)c;
      f->normal.buf_size++;
      f->normal.flags &= ~PACKFILE_FLAG_EOF;
      return (unsigned char)c;
   }
}



static long normal_fread(void *p, long n, void *_f)
{
   PACKFILE *f = _f;
   unsigned char *cp = (unsigned char *)p;
   long i;
   int c;

   for (i=0; i<n; i++) {
      if ((c = normal_getc(f)) == EOF)
	 break;

      *(cp++) = c;
   }

   return i;
}



static int normal_putc(int c, void *_f)
{
   PACKFILE *f = _f;

   if (f->normal.buf_size + 1 >= F_BUF_SIZE) {
      if (normal_flush_buffer(f, 0))
	 return EOF;
   }

   f->normal.buf_size++;
   return (*(f->normal.buf_pos++) = c);
}



static long normal_fwrite(const void *p, long n, void *_f)
{
   PACKFILE *f = _f;
   const unsigned char *cp = (const unsigned char *)p;
   long i;

   for (i=0; i<n; i++) {
      if (normal_putc(*cp++, f) == EOF)
	 break;
   }

   return i;
}



#define MIN(x,y)     (((x) < (y)) ? (x) : (y))
#define MAX(x,y)     (((x) > (y)) ? (x) : (y))

static int normal_fseek(void *_f, int offset)
{
   PACKFILE *f = _f;
   int i;

   if (f->normal.flags & PACKFILE_FLAG_WRITE)
      return -1;

   /* skip forward through the buffer */
   if (f->normal.buf_size > 0) {
      i = MIN(offset, f->normal.buf_size);
      f->normal.buf_size -= i;
      f->normal.buf_pos += i;
      offset -= i;
      if ((f->normal.buf_size <= 0) && normal_no_more_input(f))
	 f->normal.flags |= PACKFILE_FLAG_EOF;
   }

   /* need to seek some more? */
   if (offset > 0) {
      i = MIN(offset, f->normal.todo);

      if ((f->normal.flags & PACKFILE_FLAG_PACK) || (f->normal.passpos)) {
	 /* for compressed or encrypted files, we just have to read through the data */
	 while (i > 0) {
	    pack_getc(f);
	    i--;
	 }
      }
      else {
	 if (f->normal.parent) {
	    /* pass the seek request on to the parent file */
	    pack_fseek(f->normal.parent, i);
	 }
	 else {
	    /* do a real seek */
	    al_fseek(f->normal.fp, i, SEEK_CUR);
	 }
	 f->normal.todo -= i;
	 if (normal_no_more_input(f))
	    f->normal.flags |= PACKFILE_FLAG_EOF;
      }
   }

   return 0;
}



static int normal_feof(void *_f)
{
   PACKFILE *f = _f;

   return (f->normal.flags & PACKFILE_FLAG_EOF);
}



static int normal_ferror(void *_f)
{
   PACKFILE *f = _f;

   return (f->normal.flags & PACKFILE_FLAG_ERROR);
}



/* normal_refill_buffer:
 *  Refills the read buffer. The file must have been opened in read mode,
 *  and the buffer must be empty.
 */
static int normal_refill_buffer(PACKFILE *f)
{
   int i, sz, done, offset;

   if (f->normal.flags & PACKFILE_FLAG_EOF)
      return EOF;

   if (normal_no_more_input(f)) {
      f->normal.flags |= PACKFILE_FLAG_EOF;
      return EOF;
   }

   if (f->normal.parent) {
      if (f->normal.flags & PACKFILE_FLAG_PACK) {
	 f->normal.buf_size = lzss_read(f->normal.parent, f->normal.unpack_data, MIN(F_BUF_SIZE, f->normal.todo), f->normal.buf);
      }
      else {
	 f->normal.buf_size = pack_fread(f->normal.buf, MIN(F_BUF_SIZE, f->normal.todo), f->normal.parent);
      } 
      if (f->normal.parent->normal.flags & PACKFILE_FLAG_EOF)
	 f->normal.todo = 0;
      if (f->normal.parent->normal.flags & PACKFILE_FLAG_ERROR)
	 goto Error;
   }
   else {
      f->normal.buf_size = MIN(F_BUF_SIZE, f->normal.todo);

      offset = al_fseek(f->normal.fp, 0, SEEK_CUR);
      done = 0;

      errno = 0;
      sz = al_fread(f->normal.fp, f->normal.buf, f->normal.buf_size);

      while (sz+done < f->normal.buf_size) {
	 if ((sz < 0) && ((errno != EINTR) && (errno != EAGAIN)))
	    goto Error;

	 if (sz > 0)
	    done += sz;

	 al_fseek(f->normal.fp, offset+done, SEEK_SET);
         errno = 0;
	 sz = al_fread(f->normal.fp, f->normal.buf+done, f->normal.buf_size-done);
      }

      if ((f->normal.passpos) && (!(f->normal.flags & PACKFILE_FLAG_OLD_CRYPT))) {
	 for (i=0; i<f->normal.buf_size; i++) {
	    f->normal.buf[i] ^= *(f->normal.passpos++);
	    if (!*f->normal.passpos)
	       f->normal.passpos = f->normal.passdata;
	 }
      }
   }

   f->normal.todo -= f->normal.buf_size;
   f->normal.buf_pos = f->normal.buf;
   f->normal.buf_size--;
   if (f->normal.buf_size <= 0)
      if (normal_no_more_input(f))
	 f->normal.flags |= PACKFILE_FLAG_EOF;

   if (f->normal.buf_size < 0)
      return EOF;
   else
      return *(f->normal.buf_pos++);

 Error:
   f->normal.flags |= PACKFILE_FLAG_ERROR;
   return EOF;
}



/* normal_flush_buffer:
 *  Flushes a file buffer to the disk. The file must be open in write mode.
 */
static int normal_flush_buffer(PACKFILE *f, int last)
{
   int i, sz, done, offset;

   if (f->normal.buf_size > 0) {
      if (f->normal.flags & PACKFILE_FLAG_PACK) {
	 if (lzss_write(f->normal.parent, f->normal.pack_data, f->normal.buf_size, f->normal.buf, last))
	    goto Error;
      }
      else {
	 if ((f->normal.passpos) && (!(f->normal.flags & PACKFILE_FLAG_OLD_CRYPT))) {
	    for (i=0; i<f->normal.buf_size; i++) {
	       f->normal.buf[i] ^= *(f->normal.passpos++);
	       if (!*f->normal.passpos)
		  f->normal.passpos = f->normal.passdata;
	    }
	 }

	 offset = al_fseek(f->normal.fp, 0, SEEK_CUR);
	 done = 0;

	 errno = 0;
	 sz = al_fwrite(f->normal.fp, f->normal.buf, f->normal.buf_size);

	 while (sz+done < f->normal.buf_size) {
	    if ((sz < 0) && ((errno != EINTR) && (errno != EAGAIN)))
	       goto Error;

	    if (sz > 0)
	       done += sz;

	    al_fseek(f->normal.fp, offset+done, SEEK_SET);
	    errno = 0;
	    sz = al_fwrite(f->normal.fp, f->normal.buf+done, f->normal.buf_size-done);
	 }
      }
      f->normal.todo += f->normal.buf_size;
   }

   f->normal.buf_pos = f->normal.buf;
   f->normal.buf_size = 0;
   return 0;

 Error:
   f->normal.flags |= PACKFILE_FLAG_ERROR;
   return EOF;
}

void * fi_pack_fopen(const char *path, const char *mode)
{
	return pack_fopen(path, mode);
}

bool fi_pack_fclose(ALLEGRO_FILE *f)
{
	pack_fclose(al_get_file_userdata(f));
	return true;
}

size_t fi_pack_fread(ALLEGRO_FILE *f, void *ptr, size_t size)
{
	return pack_fread(ptr, size, al_get_file_userdata(f));
}

size_t fi_pack_fwrite(ALLEGRO_FILE *f, const void *ptr, size_t size)
{
	return pack_fwrite(ptr, size, al_get_file_userdata(f));
}

bool fi_pack_fflush(ALLEGRO_FILE *f)
{
	return true;
}

int64_t fi_pack_ftell(ALLEGRO_FILE *f)
{
	PACKFILE * fp = al_get_file_userdata(f);
	
	return fp->normal.pos;
}

bool fi_pack_fseek(ALLEGRO_FILE *f, int64_t offset, int whence)
{
	return pack_fseek(al_get_file_userdata(f), offset) ? true : false;
}

bool fi_pack_feof(ALLEGRO_FILE *f)
{
	return pack_feof(al_get_file_userdata(f)) ? true : false;
}

int fi_pack_ferror(ALLEGRO_FILE *f)
{
	return pack_ferror(al_get_file_userdata(f));
}

void fi_pack_fclearerr(ALLEGRO_FILE *f)
{
}

int fi_pack_fungetc(ALLEGRO_FILE *f, int c)
{
	return 0;
}

off_t fi_pack_fsize(ALLEGRO_FILE *f)
{
	return -1;
}
