/*         ______   ___    ___
 *        /\  _  \ /\_ \  /\_ \
 *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___
 *         \ \  __ \ \ \ \  \ \ \   /'__`\ /'_ `\/\`'__\/ __`\
 *          \ \ \/\ \ \_\ \_ \_\ \_/\  __//\ \L\ \ \ \//\ \L\ \
 *           \ \_\ \_\/\____\/\____\ \____\ \____ \ \_\\ \____/
 *            \/_/\/_/\/____/\/____/\/____/\/___L\ \/_/ \/___/
 *                                           /\____/
 *                                           \_/__/
 *
 *      File I/O.
 *
 *      By Shawn Hargreaves.
 *
 *      See readme.txt for copyright information.
 */


#ifndef ALLEGRO_FILE_H
#define ALLEGRO_FILE_H

#include <allegro5/allegro.h>

#ifdef __cplusplus
   extern "C" {
#endif

#ifndef EOF
   #define EOF    (-1)
#endif

#define F_READ          "r"
#define F_WRITE         "w"
#define F_READ_PACKED   "rp"
#define F_WRITE_PACKED  "wp"
#define F_WRITE_NOPACK  "w!"

#define F_BUF_SIZE      4096           /* 4K buffer for caching data */
#define F_PACK_MAGIC    0x736C6821L    /* magic number for packed files */
#define F_NOPACK_MAGIC  0x736C682EL    /* magic number for autodetect */
#define F_EXE_MAGIC     0x736C682BL    /* magic number for appended data */

#define PACKFILE_FLAG_WRITE      1     /* the file is being written */
#define PACKFILE_FLAG_PACK       2     /* data is compressed */
#define PACKFILE_FLAG_CHUNK      4     /* file is a sub-chunk */
#define PACKFILE_FLAG_EOF        8     /* reached the end-of-file */
#define PACKFILE_FLAG_ERROR      16    /* an error has occurred */
#define PACKFILE_FLAG_OLD_CRYPT  32    /* backward compatibility mode */
#define PACKFILE_FLAG_EXEDAT     64    /* reading from our executable */


typedef struct PACKFILE_VTABLE PACKFILE_VTABLE;
typedef struct PACKFILE PACKFILE;

struct LZSS_PACK_DATA;
struct LZSS_UNPACK_DATA;


struct _al_normal_packfile_details
{
   ALLEGRO_FILE * fp;                           /* DOS file handle */
   int flags;                          /* PACKFILE_FLAG_* constants */
   unsigned char *buf_pos;             /* position in buffer */
   int buf_size;                       /* number of bytes in the buffer */
   long todo;                          /* number of bytes still on the disk */
   struct PACKFILE *parent;            /* nested, parent file */
   struct LZSS_PACK_DATA *pack_data;   /* for LZSS compression */
   struct LZSS_UNPACK_DATA *unpack_data; /* for LZSS decompression */
   char *filename;                     /* name of the file */
   char *mode;
   char *passdata;                     /* encryption key data */
   char *passpos;                      /* current key position */
   unsigned char buf[F_BUF_SIZE];      /* the actual data buffer */
   int pos;                            /* file position as if the file were uncompressed */
};


struct PACKFILE                           /* our very own FILE structure... */
{
   const PACKFILE_VTABLE *vtable;
   void *userdata;
   int is_normal_packfile;

   /* The following is only to be used for the "normal" PACKFILE vtable,
    * i.e. what is implemented by Allegro itself. If is_normal_packfile is
    * false then the following is not even allocated. This must be the last
    * member in the structure.
    */
   struct _al_normal_packfile_details normal;
};


struct PACKFILE_VTABLE
{
   int(*pf_fclose)(void *);
   int(*pf_getc)(void *);
   int(*pf_ungetc)(int, void *);
   long(*pf_fread)(void *, long, void *);
   int(*pf_putc)(int, void *);
   long(*pf_fwrite)(const void *, long, void *);
   int(*pf_fseek)(void *, int);
   int(*pf_feof)(void *);
   int(*pf_ferror)(void *);
};


#define uconvert_tofilename(s, buf)      uconvert(s, U_CURRENT, buf, get_filename_encoding(), sizeof(buf))

void packfile_password(const char *password);
PACKFILE * pack_fopen(const char *filename, const char *mode);
PACKFILE * pack_fopen_vtable(const PACKFILE_VTABLE *vtable, void *userdata);
int pack_fclose(PACKFILE *f);
int pack_fseek(PACKFILE *f, int offset);
PACKFILE * pack_fopen_chunk(PACKFILE *f, int pack);
PACKFILE * pack_fclose_chunk(PACKFILE *f);
int pack_getc(PACKFILE *f);
int pack_putc(int c, PACKFILE *f);
int pack_feof(PACKFILE *f);
int pack_ferror(PACKFILE *f);
int pack_igetw(PACKFILE *f);
long pack_igetl(PACKFILE *f);
int pack_iputw(int w, PACKFILE *f);
long pack_iputl(long l, PACKFILE *f);
int pack_mgetw(PACKFILE *f);
long pack_mgetl(PACKFILE *f);
int pack_mputw(int w, PACKFILE *f);
long pack_mputl(long l, PACKFILE *f);
long pack_fread(void *p, long n, PACKFILE *f);
long pack_fwrite(const void *p, long n, PACKFILE *f);
int pack_ungetc(int c, PACKFILE *f);
char * pack_fgets(char *p, int max, PACKFILE *f);
int pack_fputs(const char *p, PACKFILE *f);
void * pack_get_userdata(PACKFILE *f);

void * fi_pack_fopen(const char *path, const char *mode);
bool fi_pack_fclose(ALLEGRO_FILE *f);
size_t fi_pack_fread(ALLEGRO_FILE *f, void *ptr, size_t size);
size_t fi_pack_fwrite(ALLEGRO_FILE *f, const void *ptr, size_t size);
bool fi_pack_fflush(ALLEGRO_FILE *f);
int64_t fi_pack_ftell(ALLEGRO_FILE *f);
bool fi_pack_fseek(ALLEGRO_FILE *f, int64_t offset, int whence);
bool fi_pack_feof(ALLEGRO_FILE *f);
int fi_pack_ferror(ALLEGRO_FILE *f);
void fi_pack_fclearerr(ALLEGRO_FILE *f);
int fi_pack_fungetc(ALLEGRO_FILE *f, int c);
off_t fi_pack_fsize(ALLEGRO_FILE *f);

#ifdef __cplusplus
   }
#endif

#endif          /* ifndef ALLEGRO_FILE_H */


