#include "t3f/t3f.h"
#include "t3f/file.h"

#include "asset.h"

static bool t3f_save_color_f(ALLEGRO_FILE * fp, ALLEGRO_COLOR * color)
{
	unsigned char r, g, b, a;

	al_unmap_rgba(*color, &r, &g, &b, &a);
	if(al_fputc(fp, r) != r)
	{
		return false;
	}
	if(al_fputc(fp, g) != g)
	{
		return false;
	}
	if(al_fputc(fp, b) != b)
	{
		return false;
	}
	if(al_fputc(fp, a) != a)
	{
		return false;
	}

	return true;
}

static bool t3f_load_color_f(ALLEGRO_FILE * fp, ALLEGRO_COLOR * color)
{
	int r, g, b, a;

	r = al_fgetc(fp);
	if(r == EOF)
	{
		return false;
	}
	g = al_fgetc(fp);
	if(g == EOF)
	{
		return false;
	}
	b = al_fgetc(fp);
	if(b == EOF)
	{
		return false;
	}
	a = al_fgetc(fp);
	if(a == EOF)
	{
		return false;
	}
	*color = al_map_rgba(r, g, b, a);

	return true;
}

CRSED_ASSET * crsed_create_asset(void)
{
    CRSED_ASSET * ap;

    ap = malloc(sizeof(CRSED_ASSET));
    if(ap)
    {
        memset(ap, 0, sizeof(CRSED_ASSET));
        ap->default_color = t3f_color_white;
    }
    return ap;
}

static ALLEGRO_BITMAP * create_bordered_bitmap(ALLEGRO_BITMAP * bp)
{
    ALLEGRO_BITMAP * bordered_bp;
    ALLEGRO_STATE old_state;
    ALLEGRO_TRANSFORM identity;

    if(bp)
    {
        bordered_bp = al_create_bitmap(al_get_bitmap_width(bp) + 2, al_get_bitmap_height(bp) + 2);
        if(bordered_bp)
        {
            al_store_state(&old_state, ALLEGRO_STATE_BLENDER | ALLEGRO_STATE_TRANSFORM | ALLEGRO_STATE_TARGET_BITMAP);
            al_identity_transform(&identity);
            al_set_target_bitmap(bordered_bp);
            al_use_transform(&identity);
            al_set_blender(ALLEGRO_ADD, ALLEGRO_ONE, ALLEGRO_ZERO);
            al_clear_to_color(al_map_rgba_f(0.0, 0.0, 0.0, 0.0));
            al_draw_bitmap(bp, 1, 1, 0);
            al_restore_state(&old_state);
            return bordered_bp;
        }
    }
    return NULL;
}

CRSED_ASSET * crsed_load_asset_f(ALLEGRO_FILE * fp, const char * fn)
{
    CRSED_ASSET * ap;
    ALLEGRO_BITMAP * bp;
    char header[16] = {0};
    int i, l;

    ap = crsed_create_asset();
    if(ap)
    {
        al_fread(fp, header, 16);
        if(strcmp(header, "vd_asset"))
        {
            crsed_destroy_asset(ap);
            return NULL;
        }
        ap->bitmaps = al_fgetc(fp);
        for(i = 0; i < ap->bitmaps; i++)
        {
            ap->source_bitmap[i] = t3f_load_bitmap_f(fp);
            if(!ap->source_bitmap[i])
            {
                crsed_destroy_asset(ap);
                return NULL;
            }
        }
        ap->type = al_fread32le(fp);

        /* generate transparent border for non-repeating textures */
        if(ap->type == CRSED_ASSET_PATTERN)
        {
            for(i = 0; i < ap->bitmaps; i++)
            {
                ap->bitmap[i] = al_create_sub_bitmap(ap->source_bitmap[i], 0, 0, al_get_bitmap_width(ap->source_bitmap[i]), al_get_bitmap_height(ap->source_bitmap[i]));
            }
        }
        else
        {
            for(i = 0; i < ap->bitmaps; i++)
            {
                bp = ap->source_bitmap[i];
                ap->source_bitmap[i] = create_bordered_bitmap(bp);
                al_destroy_bitmap(bp);
                if(!ap->source_bitmap[i])
                {
                    crsed_destroy_asset(ap);
                    return NULL;
                }
                ap->bitmap[i] = al_create_sub_bitmap(ap->source_bitmap[i], 1, 1, al_get_bitmap_width(ap->source_bitmap[i]) - 2, al_get_bitmap_height(ap->source_bitmap[i]) - 2);
            }
        }

        /* read object-specific data */
        switch(ap->type)
        {
            case CRSED_ASSET_OBJECT:
            {
                ap->poly.vertices = al_fread32le(fp);
                for(i = 0; i < ap->poly.vertices; i++)
                {
                    ap->poly.vertex[i].x = al_fixtof(al_fread32le(fp));
                    ap->poly.vertex[i].y = al_fixtof(al_fread32le(fp));
                }
                ap->r = al_get_bitmap_width(ap->bitmap[0]) / 2;
                break;
            }
            case CRSED_ASSET_WALL:
            {
                ap->poly.vertices = al_fread32le(fp);
                for(i = 0; i < ap->poly.vertices; i++)
                {
                    ap->poly.vertex[i].x = al_fixtof(al_fread32le(fp));
                    ap->poly.vertex[i].y = al_fixtof(al_fread32le(fp));
                }
                break;
            }
            case CRSED_ASSET_VERT:
            {
                ap->vert_type = al_fread32le(fp);
                ap->x = t3f_fread_float(fp);
                ap->y = t3f_fread_float(fp);
                ap->r = t3f_fread_float(fp);
                break;
            }
        }
        t3f_load_color_f(fp, &ap->default_color);

        /* read meta data tags */
        l = al_fread32le(fp);
        al_fread(fp, ap->name, l);
        l = al_fread32le(fp);
        al_fread(fp, ap->author, l);
        l = al_fread32le(fp);
        al_fread(fp, ap->comment, l);
        return ap;
    }
    return NULL;
}

CRSED_ASSET * crsed_load_asset(const char * fn)
{
    CRSED_ASSET * ap = NULL;
    ALLEGRO_FILE * fp;

    if(fn[0] == '#')
    {
        ap = crsed_create_asset();
        if(ap)
        {
            memset(ap, 0, sizeof(CRSED_ASSET));
            ap->type = CRSED_ASSET_VERT;
            ap->vert_type = atoi(&fn[1]);
            ap->default_color = t3f_color_white;
            strcpy(ap->name, fn);
        }
    }
    else
    {
        fp = al_fopen(fn, "rb");
        if(fp)
        {
            ap = crsed_load_asset_f(fp, fn);
            al_fclose(fp);
        }
    }
    return ap;
}

bool crsed_save_asset_f(CRSED_ASSET * ap, ALLEGRO_FILE * fp)
{
    bool ret = true;
    char header[16] = {'v', 'd', '_', 'a', 's', 's', 'e', 't'};
    int i;

    al_fwrite(fp, header, 16);
    al_fputc(fp, ap->bitmaps);
    for(i = 0; i < ap->bitmaps; i++)
    {
        if(!t3f_save_bitmap_f(fp, ap->bitmap[i]))
        {
            return false;
        }
    }
    al_fwrite32le(fp, ap->type);
    switch(ap->type)
    {
        case CRSED_ASSET_OBJECT:
        {
            al_fwrite32le(fp, ap->poly.vertices);
            for(i = 0; i < ap->poly.vertices; i++)
            {
                al_fwrite32le(fp, al_ftofix(ap->poly.vertex[i].x));
                al_fwrite32le(fp, al_ftofix(ap->poly.vertex[i].y));
            }
            break;
        }
        case CRSED_ASSET_WALL:
        {
            al_fwrite32le(fp, ap->poly.vertices);
            for(i = 0; i < ap->poly.vertices; i++)
            {
                al_fwrite32le(fp, al_ftofix(ap->poly.vertex[i].x));
                al_fwrite32le(fp, al_ftofix(ap->poly.vertex[i].y));
            }
            break;
        }
        case CRSED_ASSET_VERT:
        {
            al_fwrite32le(fp, ap->vert_type);
            t3f_fwrite_float(fp, ap->x);
            t3f_fwrite_float(fp, ap->y);
            t3f_fwrite_float(fp, ap->r);
            break;
        }
    }
    t3f_save_color_f(fp, &ap->default_color);

    al_fwrite32le(fp, strlen(ap->name));
    al_fwrite(fp, ap->name, strlen(ap->name));
    al_fwrite32le(fp, strlen(ap->author));
    al_fwrite(fp, ap->author, strlen(ap->author));
    al_fwrite32le(fp, strlen(ap->comment));
    al_fwrite(fp, ap->comment, strlen(ap->comment));
    return ret;
}

bool crsed_save_asset(CRSED_ASSET * ap, const char * fn)
{
    ALLEGRO_FILE * fp;
    bool ret = false;

    fp = al_fopen(fn, "wb");
    if(fp)
    {
        ret = crsed_save_asset_f(ap, fp);
        al_fclose(fp);
    }
    return ret;
}

void crsed_destroy_asset(CRSED_ASSET * ap)
{
    int i;

    for(i = 0; i < ap->bitmaps; i++)
    {
        if(ap->bitmap[i])
        {
            t3f_destroy_resource(ap->bitmap[i]);
            t3f_destroy_resource(ap->source_bitmap[i]);
        }
    }
    free(ap);
}

CRSED_ASSET_INSTANCE * crsed_create_asset_instance(int asset, ALLEGRO_COLOR color, float x, float y, float scale, float angle)
{
    CRSED_ASSET_INSTANCE * aip;

    aip = malloc(sizeof(CRSED_ASSET_INSTANCE));
    if(aip)
    {
        memset(aip, 0, sizeof(CRSED_ASSET_INSTANCE));
        aip->asset = asset;
        aip->x = x;
        aip->y = y;
        aip->scale = scale;
        aip->angle = angle;
        aip->color = color;
    }
    return aip;
}

CRSED_ASSET_INSTANCE * crsed_load_asset_instance_f(ALLEGRO_FILE * fp)
{
    CRSED_ASSET_INSTANCE * ap;
    int i;

    ap = crsed_create_asset_instance(0, t3f_color_white, 0, 0, 0, 0);
    if(ap)
    {
        ap->asset = al_fread32le(fp);
        switch(ap->asset_type)
        {
            case CRSED_ASSET_BACKGROUND:
            case CRSED_ASSET_OBJECT:
            case CRSED_ASSET_VERT:
            {
                ap->x = al_fixtof(al_fread32le(fp));
                ap->y = al_fixtof(al_fread32le(fp));
                ap->scale = al_fixtof(al_fread32le(fp));
                ap->angle = al_fixtof(al_fread32le(fp));
                t3f_load_color_f(fp, &ap->color);
                break;
            }
            case CRSED_ASSET_WALL:
            {
                ap->vertices = al_fread32le(fp);
                for(i = 0; i < ap->vertices; i++)
                {
                    ap->vertex[i].x = al_fixtof(al_fread32le(fp));
                    ap->vertex[i].y = al_fixtof(al_fread32le(fp));
                    t3f_load_color_f(fp, &ap->vertex[i].color);
                }
                break;
            }
            case CRSED_ASSET_PATTERN:
            {
                ap->x = t3f_fread_float(fp);
                ap->y = t3f_fread_float(fp);
                ap->scale = al_fixtof(al_fread32le(fp));
                ap->angle = al_fixtof(al_fread32le(fp));
                ap->vertices = al_fread32le(fp);
                for(i = 0; i < ap->vertices; i++)
                {
                    ap->vertex[i].x = al_fixtof(al_fread32le(fp));
                    ap->vertex[i].y = al_fixtof(al_fread32le(fp));
                    t3f_load_color_f(fp, &ap->vertex[i].color);
                }
                break;
            }
        }
    }
    return ap;
}

bool crsed_save_asset_instance_f(CRSED_ASSET_INSTANCE * ap, ALLEGRO_FILE * fp)
{
    int i;

    al_fwrite32le(fp, ap->asset);
    switch(ap->asset_type)
    {
        case CRSED_ASSET_BACKGROUND:
        case CRSED_ASSET_OBJECT:
        case CRSED_ASSET_VERT:
        {
            al_fwrite32le(fp, al_ftofix(ap->x));
            al_fwrite32le(fp, al_ftofix(ap->y));
            al_fwrite32le(fp, al_ftofix(ap->scale));
            al_fwrite32le(fp, al_ftofix(ap->angle));
            t3f_save_color_f(fp, &ap->color);
            break;
        }
        case CRSED_ASSET_WALL:
        {
            al_fwrite32le(fp, ap->vertices);
            for(i = 0; i < ap->vertices; i++)
            {
                al_fwrite32le(fp, al_ftofix(ap->vertex[i].x));
                al_fwrite32le(fp, al_ftofix(ap->vertex[i].y));
                t3f_save_color_f(fp, &ap->color);
            }
            break;
        }
        case CRSED_ASSET_PATTERN:
        {
            t3f_fwrite_float(fp, ap->x);
            t3f_fwrite_float(fp, ap->x);
            t3f_fwrite_float(fp, ap->scale);
            t3f_fwrite_float(fp, ap->angle);
            al_fwrite32le(fp, ap->vertices);
            for(i = 0; i < ap->vertices; i++)
            {
                al_fwrite32le(fp, al_ftofix(ap->vertex[i].x));
                al_fwrite32le(fp, al_ftofix(ap->vertex[i].y));
                t3f_save_color_f(fp, &ap->color);
            }
            break;
        }
    }
    return true;
}

void crsed_destroy_asset_instance(CRSED_ASSET_INSTANCE * ap)
{
    free(ap);
}

static void crsed_get_pattern_uv(float cx, float cy, float dx, float dy, float angle, float scale, float * ox, float * oy)
{
    if(ox)
    {
        *ox = dx - cx;
    }
    if(oy)
    {
        *oy = dy - cy;
    }
}

typedef struct
{

    float vertex[1024];
    int vertex_counts[64];

    ALLEGRO_VERTEX t_vertex[512];
    int t_vertices;

} CRSED_TRIANGLE_DATA;

static int crsed_count_asset_instance_polygons(CRSED_ASSET_INSTANCE * asset_instance)
{
    int i;
    int polygons = 0;
    int start_vertex = 0;
    int current_vertices = 0;

    if(asset_instance->vertices < 3)
    {
        return 0;
    }

    /* start a new polygon when we get a vertex that sits on top of the first vertex */
    for(i = 0; i < asset_instance->vertices; i++)
    {
        current_vertices++;
        if(i != start_vertex && fabs(asset_instance->vertex[i].x - asset_instance->vertex[start_vertex].x) < 0.1 && fabs(asset_instance->vertex[i].y - asset_instance->vertex[start_vertex].y) < 0.1)
        {
            polygons++;
            current_vertices = 0;
            start_vertex = i + 1;
        }
    }
    if(current_vertices >= 3)
    {
        polygons++;
    }

    return polygons;
}

static int crsed_get_asset_instance_poly_vertex_count(CRSED_ASSET_INSTANCE * asset_instance, int poly)
{
    int i;
    int polygons = 0;
    int start_vertex = 0;
    int current_vertices = 0;

    /* start a new polygon when we get a vertex that sits on top of the first vertex */
    for(i = 0; i < asset_instance->vertices; i++)
    {
        current_vertices++;
        if(i != start_vertex && fabs(asset_instance->vertex[i].x - asset_instance->vertex[start_vertex].x) < 0.1 && fabs(asset_instance->vertex[i].y - asset_instance->vertex[start_vertex].y) < 0.1)
        {
            if(poly == polygons)
            {
                return current_vertices;
            }
            polygons++;
            current_vertices = 0;
            start_vertex = i + 1;
        }
    }
    if(poly == polygons)
    {
        return current_vertices;
    }
    polygons++;

    return 0;
}

static void crsed_emit_triangle(int p1, int p2, int p3, void * data)
{
    CRSED_TRIANGLE_DATA * triangle_data = (CRSED_TRIANGLE_DATA *)data;

    triangle_data->t_vertex[triangle_data->t_vertices].x = triangle_data->vertex[p1 * 2];
    triangle_data->t_vertex[triangle_data->t_vertices].y = triangle_data->vertex[p1 * 2 + 1];
    triangle_data->t_vertices++;
    triangle_data->t_vertex[triangle_data->t_vertices].x = triangle_data->vertex[p2 * 2];
    triangle_data->t_vertex[triangle_data->t_vertices].y = triangle_data->vertex[p2 * 2 + 1];
    triangle_data->t_vertices++;
    triangle_data->t_vertex[triangle_data->t_vertices].x = triangle_data->vertex[p3 * 2];
    triangle_data->t_vertex[triangle_data->t_vertices].y = triangle_data->vertex[p3 * 2 + 1];
    triangle_data->t_vertices++;
}

static bool is_joint(CRSED_ASSET_VERTEX * v, int vc, int i)
{
	if(t3f_distance(v[0].x, v[0].y, v[vc - 1].x, v[vc - 1].y) < 0.1)
	{
		return true;
	}
	else if(i > 0 && i < vc - 1)
	{
		return true;
	}
	return false;
}

static float afabs(float a)
{
    if(a < 0)
    {
        return a + ALLEGRO_PI * 2.0;
    }
    return a;
}

static bool vertex_equal(float x1, float y1, float x2, float y2)
{
	if(fabs(x2 - x1) < 0.1 && fabs(y2 - y1) < 0.1)
	{
		return true;
	}
	return false;
}

static float get_joint_extrusion(CRSED_ASSET * ap, CRSED_ASSET_VERTEX * v, int vc, int i)
{
	float s;
	float a1, a2, angle, s_angle;
	float x1, y1, x2, y2;
	float r = al_get_bitmap_width(ap->source_bitmap[0]) / 2;
	float l, sl;
    float dir = 1.0;
    int i_prev, i_next;

	if(is_joint(v, vc, i))
	{
        i_prev = i - 1;
        if(i_prev < 0)
        {
            i_prev = vc - 2;
        }
        if(vertex_equal(v[0].x, v[0].y, v[i].x, v[i].y))
        {
            i = 0;
        }
        i_next = i + 1;
		a1 = afabs(atan2(v[i].y - v[i_prev].y, v[i].x - v[i_prev].x));
		a2 = afabs(atan2(v[i_next].y - v[i].y, v[i_next].x - v[i].x));
		angle = ALLEGRO_PI - afabs(a2 - a1);
        if(angle < 0.0)
        {
            dir = -1.0;
        }
		s_angle = afabs(ALLEGRO_PI - angle) / 2.0;
		x1 = v[i].x - cos(a1 + ALLEGRO_PI / 2.0) * r;
		y1 = v[i].y - sin(a1 + ALLEGRO_PI / 2.0) * r;
		x2 = v[i].x - cos(a2 + ALLEGRO_PI / 2.0) * r;
		y2 = v[i].y - sin(a2 + ALLEGRO_PI / 2.0) * r;
		l = t3f_distance(x1, y1, x2, y2);
		s = fabs(l / sin(angle));
		sl = s * sin(s_angle);
		return sl * dir;
	}
	return 0.0;
}

static void crsed_draw_prim_asset_segment(CRSED_ASSET * ap, ALLEGRO_COLOR color, CRSED_ASSET_VERTEX * v, int vc, int segment)
{
	ALLEGRO_VERTEX ev[6]; // extrapolate our line to a rectangle, filling this array
	float angle, o_angle; // angle of the segment, so we can get perpendicular lines for the extrapolation
	int v_index, v_index_next;
	float r = al_get_bitmap_width(ap->source_bitmap[0]) / 2;

	v_index = segment;
    v_index_next = v_index + 1;

	o_angle = atan2(v[v_index_next].y - v[v_index].y, v[v_index_next].x - v[v_index].x);
	angle = o_angle + ALLEGRO_PI / 2.0; // get perpendicular angle

	ev[0].x = v[v_index].x - cos(angle) * r - cos(o_angle) * get_joint_extrusion(ap, v, vc, v_index);
	ev[0].y = v[v_index].y - sin(angle) * r - sin(o_angle) * get_joint_extrusion(ap, v, vc, v_index);
	ev[0].z = 0.0;
	ev[0].color = color;
    ev[0].u = 0.0;
    ev[0].v = al_get_bitmap_height(ap->source_bitmap[0]) / 2;

	ev[1].x = v[v_index].x + cos(angle) * r + cos(o_angle) * get_joint_extrusion(ap, v, vc, v_index);
	ev[1].y = v[v_index].y + sin(angle) * r + sin(o_angle) * get_joint_extrusion(ap, v, vc, v_index);
	ev[1].z = 0.0;
	ev[1].color = color;
    ev[1].u = al_get_bitmap_width(ap->source_bitmap[0]);
    ev[1].v = al_get_bitmap_height(ap->source_bitmap[0]) / 2;

	ev[2].x = v[v_index_next].x + cos(angle) * r - cos(o_angle) * get_joint_extrusion(ap, v, vc, v_index + 1);
	ev[2].y = v[v_index_next].y + sin(angle) * r - sin(o_angle) * get_joint_extrusion(ap, v, vc, v_index + 1);
	ev[2].z = 0.0;
	ev[2].color = color;
    ev[2].u = al_get_bitmap_width(ap->source_bitmap[0]);
    ev[2].v = al_get_bitmap_height(ap->source_bitmap[0]) / 2;

	ev[3].x = v[v_index_next].x + cos(angle) * r - cos(o_angle) * get_joint_extrusion(ap, v, vc, v_index + 1);
	ev[3].y = v[v_index_next].y + sin(angle) * r - sin(o_angle) * get_joint_extrusion(ap, v, vc, v_index + 1);
	ev[3].z = 0.0;
	ev[3].color = color;
    ev[3].u = al_get_bitmap_width(ap->source_bitmap[0]);
    ev[3].v = al_get_bitmap_height(ap->source_bitmap[0]) / 2;

	ev[4].x = v[v_index_next].x - cos(angle) * r + cos(o_angle) * get_joint_extrusion(ap, v, vc, v_index + 1);
	ev[4].y = v[v_index_next].y - sin(angle) * r + sin(o_angle) * get_joint_extrusion(ap, v, vc, v_index + 1);
	ev[4].z = 0.0;
	ev[4].color = color;
    ev[4].u = 0.0;
    ev[4].v = al_get_bitmap_height(ap->source_bitmap[0]) / 2;

	ev[5].x = v[v_index].x - cos(angle) * r - cos(o_angle) * get_joint_extrusion(ap, v, vc, v_index);
	ev[5].y = v[v_index].y - sin(angle) * r - sin(o_angle) * get_joint_extrusion(ap, v, vc, v_index);
	ev[5].z = 0.0;
	ev[5].color = color;
    ev[5].u = 0.0;
    ev[5].v = al_get_bitmap_height(ap->source_bitmap[0]) / 2;

	al_draw_prim(ev, NULL, ap->source_bitmap[0], 0, 6, ALLEGRO_PRIM_TRIANGLE_LIST);
}

static void crsed_draw_prim_asset_cap(CRSED_ASSET * ap, ALLEGRO_COLOR color, CRSED_ASSET_VERTEX * v, int vi, float texture_v, float texture_h)
{
    ALLEGRO_VERTEX ev[6]; // extrapolate our line to a rectangle, filling this array
    CRSED_ASSET_VERTEX new_v[2];
    float angle, o_angle; // angle of the segment, so we can get perpendicular lines for the extrapolation
	float r = al_get_bitmap_width(ap->source_bitmap[0]) / 2;
    int subi = 0;

    if(vi != 0)
    {
        subi = -1;
    }
    o_angle = atan2(v[vi + 1 + subi].y - v[vi + subi].y, v[vi + 1 + subi].x - v[vi + subi].x);
	angle = o_angle + ALLEGRO_PI / 2.0; // get perpendicular angle

    if(vi == 0)
    {
        new_v[1].x = v[0].x;
        new_v[1].y = v[0].y;
        new_v[0].x = v[0].x - cos(o_angle) * (texture_h + 1.0);
        new_v[0].y = v[0].y - sin(o_angle) * (texture_h + 1.0);
    }
    else
    {
        new_v[0].x = v[vi].x;
        new_v[0].y = v[vi].y;
        new_v[1].x = v[vi].x + cos(o_angle) * (texture_h + 1.0);
        new_v[1].y = v[vi].y + sin(o_angle) * (texture_h + 1.0);
    }

    ev[0].x = new_v[1].x - cos(angle) * r;
	ev[0].y = new_v[1].y - sin(angle) * r;
	ev[0].z = 0.0;
	ev[0].color = color;
    ev[0].u = 0.0;
    ev[0].v = texture_v + texture_h + 1.0;

	ev[1].x = new_v[1].x + cos(angle) * r;
	ev[1].y = new_v[1].y + sin(angle) * r;
	ev[1].z = 0.0;
	ev[1].color = color;
    ev[1].u = al_get_bitmap_width(ap->source_bitmap[0]);
    ev[1].v = texture_v + texture_h + 1.0;

	ev[2].x = new_v[0].x + cos(angle) * r;
	ev[2].y = new_v[0].y + sin(angle) * r;
	ev[2].z = 0.0;
	ev[2].color = color;
    ev[2].u = al_get_bitmap_width(ap->source_bitmap[0]);
    ev[2].v = texture_v;

	ev[3].x = new_v[0].x + cos(angle) * r;
	ev[3].y = new_v[0].y + sin(angle) * r;
	ev[3].z = 0.0;
	ev[3].color = color;
    ev[3].u = al_get_bitmap_width(ap->source_bitmap[0]);
    ev[3].v = texture_v;

	ev[4].x = new_v[0].x - cos(angle) * r;
	ev[4].y = new_v[0].y - sin(angle) * r;
	ev[4].z = 0.0;
	ev[4].color = color;
    ev[4].u = 0.0;
    ev[4].v = texture_v;

	ev[5].x = new_v[1].x - cos(angle) * r;
	ev[5].y = new_v[1].y - sin(angle) * r;
	ev[5].z = 0.0;
	ev[5].color = color;
    ev[5].u = 0.0;
    ev[5].v = texture_v + texture_h + 1.0;

	al_draw_prim(ev, NULL, ap->source_bitmap[0], 0, 6, ALLEGRO_PRIM_TRIANGLE_LIST);
}

static void crsed_draw_prim_asset(CRSED_ASSET * ap, ALLEGRO_COLOR color, CRSED_ASSET_VERTEX * v, int vc)
{
	int i;

	for(i = 0; i < vc - 1; i++)
	{
		crsed_draw_prim_asset_segment(ap, color, v, vc, i);
	}

    /* render caps */
    if(vc >= 2 && !is_joint(v, vc, 0))
    {
        crsed_draw_prim_asset_cap(ap, color, v, 0, 0, ap->poly.vertex[0].y);
        crsed_draw_prim_asset_cap(ap, color, v, vc - 1, ap->poly.vertex[1].y + 1, al_get_bitmap_height(ap->bitmap[0]) - ap->poly.vertex[1].y);
    }
}

void crsed_render_asset_instance(CRSED_ASSET_INSTANCE * asset_instance, CRSED_ASSET * asset, float ox, float oy, int flags)
{
    CRSED_TRIANGLE_DATA triangle_data;
    float s;
    int i, v;

    switch(asset->type)
    {
        case CRSED_ASSET_BACKGROUND:
        {
            al_draw_tinted_bitmap(asset->bitmap[0], asset_instance->color, asset_instance->x - al_get_bitmap_width(asset->bitmap[0]) / 2, asset_instance->y - al_get_bitmap_height(asset->bitmap[0]) / 2, 0);
            break;
        }
        case CRSED_ASSET_OBJECT:
        {
            s = asset_instance->scale / asset->r;
            al_draw_tinted_scaled_rotated_bitmap(asset->bitmap[0], asset_instance->color, al_get_bitmap_width(asset->bitmap[0]) / 2, al_get_bitmap_height(asset->bitmap[0]) / 2, asset_instance->x, asset_instance->y, s, s, asset_instance->angle, 0);
            break;
        }
        case CRSED_ASSET_VERT:
        {
            s = asset_instance->scale / asset->r;
            if(asset->bitmaps && asset->bitmap[0])
            {
                al_draw_tinted_scaled_rotated_bitmap(asset->bitmap[0], asset_instance->color, al_get_bitmap_width(asset->bitmap[0]) / 2, al_get_bitmap_height(asset->bitmap[0]) / 2, asset_instance->x, asset_instance->y, s, s, asset_instance->angle, 0);
            }
            break;
        }
        case CRSED_ASSET_PATTERN:
        {
            if(asset_instance->vertices > 2)
            {
                /* triangulate polygon so we can render it as a triangle list */
                triangle_data.t_vertices = 0;
                v = asset_instance->vertices;
                for(i = 0; i < v; i++)
                {
                    triangle_data.vertex[i * 2] = asset_instance->vertex[i].x;
    				triangle_data.vertex[i * 2 + 1] = asset_instance->vertex[i].y;
                }
                for(i = 0; i < crsed_count_asset_instance_polygons(asset_instance); i++)
                {
                    triangle_data.vertex_counts[i] = crsed_get_asset_instance_poly_vertex_count(asset_instance, i);
                }
                triangle_data.vertex_counts[i] = 0;
                al_triangulate_polygon(triangle_data.vertex, sizeof(float) * 2, triangle_data.vertex_counts, crsed_emit_triangle, &triangle_data);

                /* fill out rest of data for triangle list vertices */
                for(i = 0; i < triangle_data.t_vertices; i++)
                {
                    triangle_data.t_vertex[i].z = 0.0;
                    crsed_get_pattern_uv(0, 0, triangle_data.t_vertex[i].x, triangle_data.t_vertex[i].y, asset_instance->angle, asset_instance->scale, &triangle_data.t_vertex[i].u, &triangle_data.t_vertex[i].v);
                    triangle_data.t_vertex[i].color = asset_instance->color;
                }

                /* draw the triangle list */
                al_draw_prim(triangle_data.t_vertex, NULL, asset->bitmap[0], 0, triangle_data.t_vertices, ALLEGRO_PRIM_TRIANGLE_LIST);
            }
            break;
        }
        case CRSED_ASSET_WALL:
        {
            crsed_draw_prim_asset(asset, asset_instance->color, asset_instance->vertex, asset_instance->vertices);
            break;
        }
    }
}
